# -*- coding: utf-8 -*-
"""Functionality to enhance checking of queue mainly during unit testing.

This module should not need to import from any other modules in
stdlib_utils.
"""
from __future__ import annotations

import multiprocessing
import multiprocessing.queues
import queue
from queue import Empty
from queue import Queue
import time
from typing import Any
from typing import List
from typing import Union


def _eventually_empty(
    should_be_empty: bool,
    the_queue: Union[
        Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
        multiprocessing.queues.Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
    ],
) -> bool:
    """Help to determine if queue is eventually empty or not."""
    start_time = time.process_time()
    while time.process_time() - start_time < 0.2:
        is_empty = the_queue.empty()
        value_to_check = is_empty
        if not should_be_empty:
            value_to_check = not value_to_check
        if value_to_check:
            return True
    return False


def is_queue_eventually_empty(
    the_queue: Union[
        Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
        multiprocessing.queues.Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
    ]
) -> bool:
    """Check if queue is empty prior to timeout occurring."""
    return _eventually_empty(True, the_queue)


def is_queue_eventually_not_empty(
    the_queue: Union[
        Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
        multiprocessing.queues.Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Any
        ],
    ]
) -> bool:
    """Check if queue is not empty prior to timeout occurring."""
    return _eventually_empty(False, the_queue)


def safe_get(the_queue: Queue[Any]) -> Any:  # pylint: disable=unsubscriptable-object
    try:
        return the_queue.get(block=True, timeout=0.02)
    except Empty:
        return None


def drain_queue(
    the_queue: Queue[Any],  # pylint: disable=unsubscriptable-object
) -> List[Any]:
    items = list()
    while not the_queue.empty():
        item = safe_get(the_queue)
        if item is not None:
            items.append(item)
    return items


class SimpleMultiprocessingQueue(multiprocessing.queues.SimpleQueue):  # type: ignore[type-arg] # noqa: F821 # Eli (3/10/20) can't figure out why SimpleQueue doesn't have type arguments defined in the stdlib(?)
    """Some additional basic functionality.

    Since SimpleQueue is not technically a class, there are some tricks to subclassing it: https://stackoverflow.com/questions/39496554/cannot-subclass-multiprocessing-queue-in-python-3-5
    """

    def __init__(self) -> None:
        ctx = multiprocessing.get_context()
        super().__init__(ctx=ctx)

    def get_nowait(self) -> Any:
        """Get value or raise error if empty."""
        if self.empty():
            raise queue.Empty()
        return self.get()

    def put_nowait(self, obj: Any) -> None:
        """Put without waiting/blocking.

        This is the only option with a SimpleQueue, but this is aliased
        to put to make the interface compatible with the regular
        multiprocessing.Queue interface.
        """
        self.put(obj)



# -*- coding: utf-8 -*-
"""Utilities for multiprocessing."""
from __future__ import annotations

import logging
import multiprocessing
from multiprocessing import Event
from multiprocessing import Process
import multiprocessing.queues
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

from .misc import get_formatted_stack_trace
from .parallelism_framework import InfiniteLoopingParallelismMixIn
from .queue_utils import SimpleMultiprocessingQueue


# pylint: disable=duplicate-code
class InfiniteProcess(InfiniteLoopingParallelismMixIn, Process):
    """Process with some enhanced functionality.

    Because of the more explict error reporting/handling during the run method, the Process.exitcode value will still be 0 when the process exits after handling an error.

    Args:
        fatal_error_reporter: set up as a queue to be thread/process safe. If any error is unhandled during run, it is fed into this queue so that calling thread can know the full details about the problem in this process.
    """

    # pylint: disable=duplicate-code

    def __init__(
        self,
        fatal_error_reporter: Union[
            SimpleMultiprocessingQueue,
            multiprocessing.queues.Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
                Any
            ],
        ],
        logging_level: int = logging.INFO,
        minimum_iteration_duration_seconds: Union[float, int] = 0.01,
    ) -> None:
        Process.__init__(self)
        InfiniteLoopingParallelismMixIn.__init__(
            self,
            fatal_error_reporter,
            logging_level,
            Event(),
            Event(),
            Event(),
            Event(),
            minimum_iteration_duration_seconds=minimum_iteration_duration_seconds,
        )

    def _report_fatal_error(self, the_err: Exception) -> None:
        formatted_stack_trace = get_formatted_stack_trace(the_err)
        reporter = self._fatal_error_reporter
        if not isinstance(
            reporter, (SimpleMultiprocessingQueue, multiprocessing.queues.Queue)
        ):
            raise NotImplementedError(
                "The error reporter for InfiniteProcess must by a SimpleMultiprocessingQueue or multiprocessing.Queue"
            )
        reporter.put_nowait((the_err, formatted_stack_trace))

    # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
    def run(  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        self,
        num_iterations: Optional[  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
            int  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        ] = None,
        perform_setup_before_loop: bool = True,
        perform_teardown_after_loop: bool = True,  # pylint: disable=duplicate-code
    ) -> None:
        # For some reason pylint freaks out if this method is only defined in the MixIn https://github.com/PyCQA/pylint/issues/1233
        # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        super().run(
            num_iterations=num_iterations,
            perform_setup_before_loop=perform_setup_before_loop,  # pylint: disable=duplicate-code
            perform_teardown_after_loop=perform_teardown_after_loop,
        )

    @staticmethod
    def log_and_raise_error_from_reporter(error_info: Tuple[Exception, str]) -> None:  # type: ignore[override] # noqa: F821 # we are not calling the super function here, we are completely overriding the type of object it accepts
        err, formatted_traceback = error_info
        logging.exception(formatted_traceback)
        raise err



# -*- coding: utf-8 -*-
"""Utilities for dealing with ports/sockets."""
import socket
import time
from typing import Union

from .exceptions import PortNotInUseError
from .exceptions import PortUnavailableError


def is_port_in_use(port: int, host: str = "127.0.0.1") -> bool:
    """Check if port is in use.

    Obtained directly from https://stackoverflow.com/questions/2470971/fast-way-to-test-if-a-port-is-in-use-using-python
    Not unit tested...not sure best way how. But other methods call it.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as temp_socket:
        return temp_socket.connect_ex((host, port)) == 0


def _confirm_port_availability(
    expected_to_be_available: bool,
    port: int,
    host: str = "127.0.0.1",
    timeout: Union[float, int] = 0,
) -> None:
    """Raise error if port not in expected state after timeout."""
    start_time = time.perf_counter()

    max_time = start_time + timeout
    while True:
        port_in_use = is_port_in_use(port, host=host)
        if not (port_in_use == expected_to_be_available):
            return
        current_time = time.perf_counter()
        if current_time > max_time:
            if expected_to_be_available:
                raise PortUnavailableError(
                    f"The port {host}:{port} was still unavailable even after waiting {timeout} seconds."
                )

            raise PortNotInUseError(
                f"The port {host}:{port} was still not in use even after waiting {timeout} seconds."
            )


def confirm_port_available(
    port: int, host: str = "127.0.0.1", timeout: Union[float, int] = 0
) -> None:
    """Raise error if port unavailable after timeout.

    Args:
        port: the port number
        host: the host to check, defaults to 127.0.0.1
        timeout: the number in seconds to wait before raising an error. defaults to 0 which is no wait, must be immediately available
    """
    _confirm_port_availability(True, port, host=host, timeout=timeout)


def confirm_port_in_use(
    port: int, host: str = "127.0.0.1", timeout: Union[float, int] = 0
) -> None:
    """Raise error if port is still open/unused after timeout.

    Args:
        port: the port number
        host: the host to check, defaults to 127.0.0.1
        timeout: the number in seconds to wait before raising an error. defaults to 0 which is no wait, must be already in use
    """
    _confirm_port_availability(False, port, host=host, timeout=timeout)



# -*- coding: utf-8 -*-
"""Functionality to enhance parallelism.

This module can import things from threading_utils,
multiprocessing_utils, and queue_utils.
"""
from __future__ import annotations

import multiprocessing
import multiprocessing.queues
from queue import Queue
from typing import Any
from typing import Dict
from typing import Union

from .multiprocessing_utils import InfiniteProcess
from .parallelism_framework import InfiniteLoopingParallelismMixIn
from .queue_utils import is_queue_eventually_not_empty
from .queue_utils import SimpleMultiprocessingQueue
from .threading_utils import InfiniteThread


def put_log_message_into_queue(
    log_level_of_this_message: int,
    the_message: Any,
    the_queue: Union[
        Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Dict[str, Any]
        ],
        SimpleMultiprocessingQueue,
        multiprocessing.queues.Queue[  # pylint: disable=unsubscriptable-object # Eli (3/12/20) not sure why pylint doesn't recognize this type annotation
            Dict[str, Any]
        ],
    ],
    log_level_threshold: int,
    pause_after_put: bool = False,
) -> None:
    """Put a log message into a queue.

    The message is only put in if the log level of the message meets the
    threshold of the queue.
    """
    if log_level_of_this_message >= log_level_threshold:
        comm_dict = {
            "communication_type": "log",
            "log_level": log_level_of_this_message,
            "message": the_message,
        }
        the_queue.put_nowait(comm_dict)
    if not isinstance(the_queue, SimpleMultiprocessingQueue) and pause_after_put:
        is_queue_eventually_not_empty(the_queue)


def invoke_process_run_and_check_errors(
    the_process: InfiniteLoopingParallelismMixIn,
    num_iterations: int = 1,
    perform_setup_before_loop: bool = False,
) -> None:
    """Call the run method of a process and raise any errors.

    This is often useful in unit testing. This should only be used on
    processes that have not been started.
    """
    the_process.run(
        num_iterations=num_iterations,
        perform_setup_before_loop=perform_setup_before_loop,
        perform_teardown_after_loop=False,
    )

    error_queue = the_process.get_fatal_error_reporter()
    is_item_in_queue = not error_queue.empty()
    if not isinstance(error_queue, SimpleMultiprocessingQueue):
        is_item_in_queue = is_queue_eventually_not_empty(error_queue)
    if is_item_in_queue:
        err_info = the_process.get_fatal_error_reporter().get_nowait()
        if isinstance(the_process, InfiniteProcess):
            if not isinstance(err_info, tuple):
                raise NotImplementedError(
                    "Errors from InfiniteProcess must be Tuple[Exception,str]"
                )
            excp, trace = err_info
            if not isinstance(excp, Exception):
                raise NotImplementedError(
                    "Errors from InfiniteProcess must be Tuple[Exception,str]"
                )
            if not isinstance(trace, str):
                raise NotImplementedError(
                    "Errors from InfiniteProcess must be Tuple[Exception,str]"
                )
            InfiniteProcess.log_and_raise_error_from_reporter((excp, trace))
        if not isinstance(err_info, Exception):

            raise NotImplementedError("Errors from InfiniteThread must be Exceptions")

        InfiniteThread.log_and_raise_error_from_reporter(err_info)



# -*- coding: utf-8 -*-
"""Misc helper utilities."""

from typing import List

if (
    6 < 9
):  # pragma: no cover # Eli (5/18/20): can't figure out a better way to stop zimports from deleting the nosec comment
    # Eli (5/18/20): need to nest it to avoid zimports deleting the comment
    from xml.etree.ElementTree import (  # nosec Eli (5/18/20): this is a false alarm from Bandit. Yes, ElementTree can parse malicious XML and should be avoided, but Element itself contains no parsing ability
        Element,
    )


class PortUnavailableError(Exception):
    pass


class PortNotInUseError(Exception):
    pass


class BlankAbsoluteResourcePathError(Exception):
    def __init__(self) -> None:
        super().__init__(
            'Supplying a blank/falsey absolute path is dangerous as Python interprets it as the current working directory, which should not be relied on as being constant. Likely fixes include using the get_current_file_abs_directory() function combined with ".." in os.path.join to create an absolute path to the location you need.'
        )


class MultipleMatchingXmlElementsError(Exception):
    def __init__(self, findall_results: List[Element]) -> None:
        super().__init__(str(findall_results))


class NoMatchingXmlElementError(Exception):
    def __init__(self, node: Element, name: str) -> None:
        super().__init__(
            f"There was no child element named '{name}' found in the XML element {node} with children {list(node)}"
        )


class LogFolderDoesNotExistError(Exception):
    pass


class LogFolderGivenWithoutFilePrefixError(Exception):
    pass


class Crc32InFileHeadDoesNotMatchExpectedValueError(Exception):
    pass


class Crc32ChecksumValidationFailureError(Exception):
    pass



# -*- coding: utf-8 -*-
"""Controlling communication with the OpalKelly FPGA Boards."""
import logging
import queue
import threading
from typing import Optional
from typing import Union

from .parallelism_framework import InfiniteLoopingParallelismMixIn


class InfiniteThread(InfiniteLoopingParallelismMixIn, threading.Thread):
    """Thread for running infinitely in the background.

    Contains some enhanced functionality for stopping.

    Args:
        fatal_error_reporter: set up as a queue to be thread/process safe. If any error is unhandled during run, it is fed into this queue so that calling thread can know the full details about the problem in this process.
    """

    def __init__(
        self,
        fatal_error_reporter: queue.Queue,  # type: ignore[type-arg] # noqa: F821 # Eli (3/10/20) can't figure out why queue.Queue doesn't have type arguments defined in the stdlib(?)
        lock: Optional[threading.Lock] = None,
        logging_level: int = logging.INFO,
        minimum_iteration_duration_seconds: Union[float, int] = 0.01,
    ) -> None:
        threading.Thread.__init__(self)
        InfiniteLoopingParallelismMixIn.__init__(
            self,
            fatal_error_reporter,
            logging_level,
            threading.Event(),
            threading.Event(),
            threading.Event(),
            threading.Event(),
            minimum_iteration_duration_seconds=minimum_iteration_duration_seconds,
        )
        self._lock = lock

    # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
    def run(  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        self,
        num_iterations: Optional[int] = None,
        perform_setup_before_loop: bool = True,  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        perform_teardown_after_loop: bool = True,
    ) -> None:  # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        # For some reason pylint freaks out if this method is only defined in the MixIn https://github.com/PyCQA/pylint/issues/1233
        # pylint: disable=duplicate-code # pylint is freaking out and requiring the method to be redefined
        super().run(
            num_iterations=num_iterations,
            perform_setup_before_loop=perform_setup_before_loop,
            perform_teardown_after_loop=perform_teardown_after_loop,
        )



# -*- coding: utf-8 -*-
"""Functions for calculating, embedding, and validating checksums."""
import struct
from typing import IO
from typing import Optional
from zlib import crc32

from .exceptions import Crc32ChecksumValidationFailureError
from .exceptions import Crc32InFileHeadDoesNotMatchExpectedValueError


def _convert_crc32_bytes_to_hex(checksum_bytes: bytes) -> str:
    checksum_int = struct.unpack(">I", checksum_bytes)[0]
    return ("%08X" % (checksum_int & 0xFFFFFFFF)).lower()


def compute_crc32_bytes_of_large_file(
    file_handle: IO[bytes], skip_first_n_bytes: int = 0
) -> bytes:
    """Calculate the CRC32 checksum in a memory-efficient manner.

    Modified from: https://stackoverflow.com/questions/1742866/compute-crc-of-file-in-python

    Args:
        file_handle: the file handle to process. Should be opened in 'rb' mode
        skip_first_n_bytes: For cases when a checksum has been written in as the first bytes of a file (e.g. in an H5 userblock), the calculation can skip those bytes

    Returns:
        The CRC32 checksum as bytes
    """
    checksum = 0
    if skip_first_n_bytes > 0:
        file_handle.read(skip_first_n_bytes)
    while True:
        itered_bytes = file_handle.read(65536)
        if not itered_bytes:
            break
        checksum = crc32(itered_bytes, checksum)
    return struct.pack(">I", checksum)


def compute_crc32_hex_of_large_file(
    file_handle: IO[bytes], skip_first_n_bytes: int = 0
) -> str:
    """Calcuates the lowercase zero-padded hex string of a file."""
    checksum_bytes = compute_crc32_bytes_of_large_file(
        file_handle, skip_first_n_bytes=skip_first_n_bytes
    )
    return _convert_crc32_bytes_to_hex(checksum_bytes)


def compute_crc32_and_write_to_file_head(  # pylint: disable=invalid-name # Eli (7/29/20): it's one character over the limit, but I can't think of a shorter way to describe it
    file_handle: IO[bytes],
) -> None:
    """Write the calculated CRC32 checksum over the first 4 bytes of the file.

    This is often used to facilitate encoding a CRC32 checksum in the Userblock of an H5 file.

    Args:
        file_handle: the file should be opened in 'rb+' mode
    """
    checksum_bytes = compute_crc32_bytes_of_large_file(
        file_handle, skip_first_n_bytes=4
    )
    file_handle.seek(0)
    file_handle.write(checksum_bytes)


def validate_file_head_crc32(
    file_handle: IO[bytes], expected_checksum: Optional[str] = None
) -> None:
    """Validate a file where the CRC32 checksum is contained in the head.

    This is often used to facilitate encoding a CRC32 checksum in the Userblock of an H5 file.

    Raises error if mismatch in checksum detected or if supplied expected checksum does not match the value found at the file head.

    Args:
        file_handle: the file should be opened in 'rb' mode
        expected_checksum: If the expected checksum is already known (from records), then this can optionally be supplied to confirm the embedded checksum in the file head matches this.
    """
    checksum_bytes_in_file_head = file_handle.read(4)
    checksum_hex_in_file_head = _convert_crc32_bytes_to_hex(checksum_bytes_in_file_head)
    if expected_checksum is not None:
        if checksum_hex_in_file_head != expected_checksum:
            raise Crc32InFileHeadDoesNotMatchExpectedValueError(
                f"The expected checksum passed in to the function was {expected_checksum}, but the checksum found at the file head is {checksum_hex_in_file_head}."
            )

    actual_checksum_bytes = compute_crc32_bytes_of_large_file(file_handle)
    if actual_checksum_bytes != checksum_bytes_in_file_head:
        actual_checksum_hex = _convert_crc32_bytes_to_hex(actual_checksum_bytes)
        raise Crc32ChecksumValidationFailureError(
            f"The checksum calculated from the file was {actual_checksum_hex}, but the checksum found at the file head is {checksum_hex_in_file_head}."
        )



# -*- coding: utf-8 -*-
"""Functionality to enhance parallelism."""
from __future__ import annotations

import logging
import multiprocessing
import multiprocessing.queues
import multiprocessing.synchronize
import queue
from statistics import stdev
import threading
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from .misc import get_formatted_stack_trace
from .misc import print_exception
from .queue_utils import is_queue_eventually_not_empty
from .queue_utils import SimpleMultiprocessingQueue


def calculate_iteration_time_ns(start_timepoint_of_iteration: int) -> int:
    return time.perf_counter_ns() - start_timepoint_of_iteration


# pylint: disable=too-many-instance-attributes
class InfiniteLoopingParallelismMixIn:
    """Mix-in for infinite looping.

    Attrs:
        num_longest_iterations: the quantity of longest iterations the object should keep track of.

    Args:
        fatal_error_reporter: a queue to report any fatal unhandled errors back to the thread that started this process
        logging_level: what threshold to use for logging events
        stop_event: When set (typically by calling .stop()), this will cause the infinite loop to exit on the next iteration
        soft_stop_event: When set (typically by calling .soft_stop()), this will cause the infinite loop to exit the next iteration that all conditionals indicating a soft_stop is possible are met (typically used to ensure all incoming tasks/queues are empty and there is nothing currently available to process)
        teardown_complete_event: After the infinite loop is exited, the _teardown_after_loop() method will be called. This event can be monitored by the parent thread to determine when the teardown has completed and the process is ready to have any needed additional clean-up performed by the parent before the parent calls .join().
        minimum_iteration_duration_seconds: In order for the process not to unnecessarily consume CPU resources while looping, the loop will sleep at the end of each iteration until this threshold duration is met.
    """

    num_longest_iterations = 5

    def __init__(
        self,
        fatal_error_reporter: Union[
            queue.Queue[str],
            multiprocessing.queues.Queue[Tuple[Exception, str]],
            SimpleMultiprocessingQueue,
        ],
        logging_level: int,
        stop_event: Union[threading.Event, multiprocessing.synchronize.Event],
        soft_stop_event: Union[threading.Event, multiprocessing.synchronize.Event],
        teardown_complete_event: Union[
            threading.Event, multiprocessing.synchronize.Event
        ],
        start_up_complete_event: Union[
            threading.Event, multiprocessing.synchronize.Event
        ],
        minimum_iteration_duration_seconds: Union[float, int] = 0.01,
    ) -> None:
        self._stop_event = stop_event
        self._soft_stop_event = soft_stop_event
        self._teardown_complete_event = teardown_complete_event
        self._start_up_complete_event = start_up_complete_event
        self._fatal_error_reporter = fatal_error_reporter
        self._process_can_be_soft_stopped = True
        self._logging_level = logging_level
        self._minimum_iteration_duration_seconds = minimum_iteration_duration_seconds
        self._idle_iteration_time_ns = 0
        self._percent_use_values: List[float] = list()
        self._longest_iterations: List[int] = list()
        self._init_performance_measurements()

    def _init_performance_measurements(self) -> None:
        # separate to make mocking easier
        self._reset_performance_measurements()

    def _reset_performance_measurements(self) -> None:
        self._start_timepoint_of_last_performance_measurement = time.perf_counter_ns()
        self._idle_iteration_time_ns = 0

    def get_start_timepoint_of_performance_measurement(self) -> int:
        return self._start_timepoint_of_last_performance_measurement

    def get_elapsed_time_since_last_performance_measurement(self) -> int:
        return (
            time.perf_counter_ns()
            - self._start_timepoint_of_last_performance_measurement
        )

    def reset_performance_tracker(self) -> Dict[str, Any]:
        """Reset performance tracking and return various metrics."""
        out_dict: Dict[str, Any] = {}
        out_dict[
            "start_timepoint_of_measurements"
        ] = self._start_timepoint_of_last_performance_measurement
        out_dict["idle_iteration_time_ns"] = self._idle_iteration_time_ns
        out_dict["percent_use"] = 100 * (
            1
            - self._idle_iteration_time_ns
            / self.get_elapsed_time_since_last_performance_measurement()
        )
        out_dict["longest_iterations"] = self._longest_iterations
        self._percent_use_values.append(out_dict["percent_use"])
        self._reset_performance_measurements()
        return out_dict

    def get_percent_use_values(self) -> List[float]:
        return self._percent_use_values

    def get_percent_use_metrics(self) -> Dict[str, float]:
        metrics = {
            "max": max(self._percent_use_values),
            "min": min(self._percent_use_values),
            "stdev": round(stdev(self._percent_use_values), 6),
            "mean": round(
                sum(self._percent_use_values) / len(self._percent_use_values), 6
            ),
        }
        return metrics

    def get_idle_time_ns(self) -> float:
        return self._idle_iteration_time_ns

    def get_minimum_iteration_duration_seconds(self) -> Union[float, int]:
        return self._minimum_iteration_duration_seconds

    def get_logging_level(self) -> int:
        return self._logging_level

    @staticmethod
    def log_and_raise_error_from_reporter(error_info: Exception) -> None:
        err = error_info
        if not isinstance(err, Exception):
            raise NotImplementedError(
                "Error in the code, this should always be an Exception."
            )
        formatted_traceback = get_formatted_stack_trace(err)
        logging.exception(formatted_traceback)
        raise err

    def get_fatal_error_reporter(
        self,
    ) -> Union[
        queue.Queue[str],
        multiprocessing.queues.Queue[Tuple[Exception, str]],
        SimpleMultiprocessingQueue,
    ]:
        return self._fatal_error_reporter

    def _report_fatal_error(self, the_err: Exception) -> None:
        self._fatal_error_reporter.put_nowait(the_err)  # type: ignore # the subclasses all have an instance of fatal error reporter. there may be a more elegant way to handle this to make mypy happy though... (Eli 2/12/20)

    def _setup_before_loop(self) -> None:
        """Perform any necessary setup prior to initiating the infinite loop.

        This can be overridden by the subclass.
        """

    def _teardown_after_loop(self) -> None:
        """Perform any necessary teardown after the infinite loop has exited.

        It's the responsibility of this method and parent process to make sure all queues get emptied before join is called.

        This can be overridden by the subclass, but the super method should always be called at the end of the subclass's implementation.
        """
        if not hasattr(self, "_teardown_complete_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _teardown_complete_event attribute."
            )
        teardown_complete_event = getattr(self, "_teardown_complete_event")

        teardown_complete_event.set()

    def run(
        self,
        num_iterations: Optional[int] = None,
        perform_setup_before_loop: bool = True,
        perform_teardown_after_loop: bool = True,
    ) -> None:
        """Run the thread.

        Args:
            num_iterations: typically used for unit testing to just execute one or a few cycles. if left as None will loop infinitely
            perform_setup_before_loop: this can be disabled when needed during unit testing
            perform_teardown_after_loop: this can be disabled when needed during unit testing

        This sets up the basic flow control and error handling for the thread.
        Subclasses should implement functionality to be executed during each
        cycle in the _commands_for_each_run_iteration method.
        """
        if num_iterations is None:
            num_iterations = -1
        completed_iterations = 0
        if perform_setup_before_loop:
            try:
                self._setup_before_loop()
            except Exception as e:  # pylint: disable=broad-except # The deliberate goal of this is to catch everything and put it into the error queue
                print_exception(e, "cf477f32-9797-417e-a157-ea6e0c4f25d1")
                self._report_fatal_error(e)
                return
        self._start_up_complete_event.set()
        while True:
            start_timepoint_of_iteration = time.perf_counter_ns()
            self._process_can_be_soft_stopped = True
            try:
                self._commands_for_each_run_iteration()
            except Exception as e:  # pylint: disable=broad-except # The deliberate goal of this is to catch everything and put it into the error queue
                print_exception(e, "88a25177-b2a1-4bbb-ba92-bf5810594a99")
                self._report_fatal_error(e)
                self.stop()
            if self.is_preparing_for_soft_stop() and self._process_can_be_soft_stopped:
                self.stop()

            if self.is_stopped():
                # Having the check for is_stopped after the first iteration of run allows easier unit testing.
                break
            completed_iterations += 1
            if completed_iterations == num_iterations:
                break
            # only decide to sleep if there are more iterations to do. this will keep unit tests executing more quickly
            self._sleep_for_idle_time_during_iteration(start_timepoint_of_iteration)
        if perform_teardown_after_loop:
            try:
                self._teardown_after_loop()
            except Exception as e:  # pylint: disable=broad-except # The deliberate goal of this is to catch everything and put it into the error queue
                print_exception(e, "bd9a8587-e79b-43cb-8ffe-0bf45740599d")
                self._report_fatal_error(e)

    def _sleep_for_idle_time_during_iteration(
        self, start_timepoint_of_iteration: int
    ) -> None:
        iteration_time_ns = calculate_iteration_time_ns(start_timepoint_of_iteration)

        longest_iterations = self._longest_iterations
        if len(longest_iterations) < 5:
            longest_iterations.append(iteration_time_ns)
        else:
            min_longest_iteration = min(longest_iterations)
            if iteration_time_ns > min_longest_iteration:
                longest_iterations[
                    longest_iterations.index(min_longest_iteration)
                ] = iteration_time_ns

        idle_time_ns = (
            int(self.get_minimum_iteration_duration_seconds() * 10 ** 9)
            - iteration_time_ns
        )
        if idle_time_ns > 0:
            self._idle_iteration_time_ns += idle_time_ns
            time.sleep(idle_time_ns / 10 ** 9)

    def _commands_for_each_run_iteration(self) -> None:
        """Execute additional commands inside the run loop."""

    def stop(self) -> None:
        """Trigger the infinite loop to break on next iteration.

        This will leave any unprocessed tasks still in incoming queues, which must be emptied before the process can be joined.

        It's the responsibility of _teardown_after_loop and parent
        process to make sure all queues get emptied before join is
        called.
        """
        if not hasattr(self, "_stop_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _stop_event attribute."
            )
        stop_event = getattr(self, "_stop_event")

        stop_event.set()

    def soft_stop(self) -> None:
        """Stop the infinite loop when the process indicates it is OK to do so.

        Typically useful for unit testing. For example waiting until all
        queued up items have been handled.

        It's the responsibility of _teardown_after_loop and parent process to make sure all queues get emptied before join is called.
        """
        if not hasattr(self, "_soft_stop_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _soft_stop_event attribute."
            )
        soft_stop_event = getattr(self, "_soft_stop_event")

        soft_stop_event.set()

    def hard_stop(self, timeout: Optional[float] = None) -> Dict[str, Any]:
        """Stop the infinite loop and drain all queues.

        Timeout can be specified (in seconds) which will override waiting for process to tear itself down.

        Items in queues will be returned in a dict
        """
        self.stop()
        self._teardown_after_loop()
        start_timepoint = time.perf_counter()
        while True:
            if timeout is not None:
                elapsed_time = time.perf_counter() - start_timepoint
                if elapsed_time > timeout:
                    break
            if self.is_teardown_complete():
                break

        item_dict = self._drain_all_queues()

        error_queue = self.get_fatal_error_reporter()
        error_items = list()
        if isinstance(error_queue, SimpleMultiprocessingQueue):
            while not error_queue.empty():
                error_items.append(error_queue.get_nowait())
        else:
            while is_queue_eventually_not_empty(error_queue):
                error_items.append(error_queue.get_nowait())

        item_dict["fatal_error_reporter"] = error_items
        return item_dict

    def _drain_all_queues(self) -> Dict[str, Any]:
        """Drain all queues of the process except the fatal_error_reporter.

        fatal_error_reporter will always be drained by hard_stop.

        This method should be overriden by subclasses implementations
        """
        # pylint:disable=no-self-use # Tanner (5/12/20): this is needed so method signature matches subclass implementation
        return dict()

    def is_start_up_complete(self) -> bool:
        """Check if the parallel instance has completed start up."""
        if not hasattr(self, "_start_up_complete_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _start_up_complete_event attribute."
            )
        start_up_complete_event = getattr(self, "_start_up_complete_event")

        is_set = start_up_complete_event.is_set()
        if not isinstance(is_set, bool):
            raise NotImplementedError(
                "The return value from this should always be a bool."
            )
        return is_set

    def is_stopped(self) -> bool:
        """Check if the parallel instance is stopped."""
        if not hasattr(self, "_stop_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _stop_event attribute."
            )
        stop_event = getattr(self, "_stop_event")

        is_set = stop_event.is_set()
        if not isinstance(is_set, bool):
            raise NotImplementedError(
                "The return value from this should always be a bool."
            )
        return is_set

    def is_preparing_for_soft_stop(self) -> bool:
        """Check if the parallel instance is preparing to soft stop."""
        if not hasattr(self, "_soft_stop_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _soft_stop_event attribute."
            )
        soft_stop_event = getattr(self, "_soft_stop_event")
        if not isinstance(
            soft_stop_event, (threading.Event, multiprocessing.synchronize.Event)
        ):
            raise NotImplementedError(
                "Classes using this mixin must have a _soft_stop_event as either a threading.Event or multiprocessing.Event"
            )
        return soft_stop_event.is_set()

    def is_teardown_complete(self) -> bool:
        """Check if the parallel instance has completed tearing itself down."""
        if not hasattr(self, "_teardown_complete_event"):
            raise NotImplementedError(
                "Classes using this mixin must have a _teardown_complete_event attribute."
            )
        teardown_complete_event = getattr(self, "_teardown_complete_event")
        if not isinstance(
            teardown_complete_event,
            (threading.Event, multiprocessing.synchronize.Event),
        ):
            raise NotImplementedError(
                "Classes using this mixin must have a _teardown_complete_event as either a threading.Event or multiprocessing.Event"
            )
        return teardown_complete_event.is_set()



# -*- coding: utf-8 -*-
"""Helpers for XML processing."""
from .exceptions import MultipleMatchingXmlElementsError
from .exceptions import NoMatchingXmlElementError

if (
    6 < 9
):  # pragma: no cover # Eli (5/18/20): can't figure out a better way to stop zimports from deleting the nosec comment # pylint: disable=duplicate-code
    # Eli (5/18/20): need to nest it to avoid zimports deleting the comment
    from xml.etree.ElementTree import (  # nosec Eli (5/18/20): this is a false alarm from Bandit. Yes, ElementTree can parse malicious XML and should be avoided, but Element itself contains no parsing ability
        Element,
    )


def find_exactly_one_xml_element(node: Element, name: str) -> Element:
    find_all_results = node.findall(name)
    if len(find_all_results) > 1:
        raise MultipleMatchingXmlElementsError(find_all_results)
    if len(find_all_results) == 0:
        raise NoMatchingXmlElementError(node, name)
    return find_all_results[0]



# -*- coding: utf-8 -*-
"""Helper utilities only requiring the standard library."""
from . import checksum
from . import loggers
from . import misc
from . import parallelism_utils
from . import ports
from .checksum import compute_crc32_and_write_to_file_head
from .checksum import compute_crc32_bytes_of_large_file
from .checksum import compute_crc32_hex_of_large_file
from .checksum import validate_file_head_crc32
from .exceptions import BlankAbsoluteResourcePathError
from .exceptions import Crc32ChecksumValidationFailureError
from .exceptions import Crc32InFileHeadDoesNotMatchExpectedValueError
from .exceptions import LogFolderDoesNotExistError
from .exceptions import LogFolderGivenWithoutFilePrefixError
from .exceptions import MultipleMatchingXmlElementsError
from .exceptions import NoMatchingXmlElementError
from .exceptions import PortNotInUseError
from .exceptions import PortUnavailableError
from .loggers import configure_logging
from .misc import create_directory_if_not_exists
from .misc import get_current_file_abs_directory
from .misc import get_current_file_abs_path
from .misc import get_formatted_stack_trace
from .misc import get_path_to_frozen_bundle
from .misc import is_frozen_as_exe
from .misc import is_system_windows
from .misc import print_exception
from .misc import resource_path
from .multiprocessing_utils import InfiniteProcess
from .parallelism_framework import InfiniteLoopingParallelismMixIn
from .parallelism_utils import invoke_process_run_and_check_errors
from .parallelism_utils import put_log_message_into_queue
from .ports import confirm_port_available
from .ports import confirm_port_in_use
from .ports import is_port_in_use
from .queue_utils import drain_queue
from .queue_utils import is_queue_eventually_empty
from .queue_utils import is_queue_eventually_not_empty
from .queue_utils import safe_get
from .queue_utils import SimpleMultiprocessingQueue
from .threading_utils import InfiniteThread
from .xml import find_exactly_one_xml_element

# from .misc import raise_alarm_signal

__all__ = [
    "configure_logging",
    "resource_path",
    "is_frozen_as_exe",
    "get_formatted_stack_trace",
    "get_path_to_frozen_bundle",
    "is_system_windows",
    "create_directory_if_not_exists",
    # "raise_alarm_signal",
    "misc",
    "checksum",
    "loggers",
    "InfiniteProcess",
    "SimpleMultiprocessingQueue",
    "invoke_process_run_and_check_errors",
    "find_exactly_one_xml_element",
    "MultipleMatchingXmlElementsError",
    "NoMatchingXmlElementError",
    "ports",
    "confirm_port_in_use",
    "confirm_port_available",
    "is_queue_eventually_empty",
    "is_queue_eventually_not_empty",
    "is_port_in_use",
    "PortUnavailableError",
    "PortNotInUseError",
    "BlankAbsoluteResourcePathError",
    "InfiniteThread",
    "InfiniteLoopingParallelismMixIn",
    "put_log_message_into_queue",
    "print_exception",
    "get_current_file_abs_directory",
    "get_current_file_abs_path",
    "parallelism_utils",
    "safe_get",
    "drain_queue",
    "LogFolderGivenWithoutFilePrefixError",
    "LogFolderDoesNotExistError",
    "compute_crc32_bytes_of_large_file",
    "compute_crc32_hex_of_large_file",
    "compute_crc32_and_write_to_file_head",
    "validate_file_head_crc32",
    "Crc32InFileHeadDoesNotMatchExpectedValueError",
    "Crc32ChecksumValidationFailureError",
]



# -*- coding: utf-8 -*-
"""Helper utilities for logging."""
import datetime
import logging
import os
import sys
import time
from typing import Optional

from .exceptions import LogFolderDoesNotExistError
from .exceptions import LogFolderGivenWithoutFilePrefixError
from .misc import create_directory_if_not_exists
from .misc import resource_path


def configure_logging(
    path_to_log_folder: Optional[str] = None,
    log_file_prefix: Optional[str] = None,
    log_level: int = logging.INFO,
) -> None:
    """Apply standard configuration to logging.

    Args:
        path_to_log_folder: optional path to an exisiting folder to use for creating log files. log_file_prefix must also be specified if this argument is not None.
        log_file_prefix: if set without path_to_log_folder specified, will write logs to file in a subfolder (logs). By default it will create a subfolder in the current working directory (if running from source) or in the path that the EXE was installed to for pyinstaller. If path_to_log_folder is specified will write logs to file in the given log folder using this as the prefix.
        log_level: set the desired logging threshold level
    """
    logging.Formatter.converter = time.gmtime  # ensure all logging timestamps are UTC
    stdout_handler = logging.StreamHandler(sys.stdout)
    handlers = [stdout_handler]
    if log_file_prefix is not None:
        log_folder: str
        if path_to_log_folder is not None:
            if not os.path.isdir(path_to_log_folder):
                raise LogFolderDoesNotExistError(path_to_log_folder)
            log_folder = path_to_log_folder
        else:
            log_folder = resource_path("logs", base_path=os.getcwd())
            create_directory_if_not_exists(log_folder)
        file_handler = logging.FileHandler(
            os.path.join(
                log_folder,
                f'{log_file_prefix}__{datetime.datetime.utcnow().strftime("%Y_%m_%d_%H%M%S")}.txt',
            )
        )
        handlers.append(file_handler)
    elif path_to_log_folder is not None:
        raise LogFolderGivenWithoutFilePrefixError()

    logging.basicConfig(
        level=log_level,
        format="[%(asctime)s UTC] %(name)s-{%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        handlers=handlers,
    )



# -*- coding: utf-8 -*-
"""Setup configuration."""
from setuptools import find_packages
from setuptools import setup


setup(
    name="stdlib_utils",
    version="0.3.0",
    description="Various utility functions and classes using only the Python standard lib",
    url="https://github.com/CuriBio/stdlib-utils",
    author="Curi Bio",
    author_email="contact@curibio.com",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-
"""Misc helper utilities."""
import inspect
import os
import sys
import traceback
from typing import Optional
from typing import Union
from uuid import UUID

from .exceptions import BlankAbsoluteResourcePathError


def get_current_file_abs_path() -> str:
    """Return the absolute path of the file that called this function."""
    return inspect.stack()[1][1]


def get_current_file_abs_directory() -> str:
    """Return the absolute directory of the file that called this function.

    The implementation cannot make a subcall to
    get_current_file_abs_path because that would add an additional item
    to the call stack.
    """
    return os.path.dirname((inspect.stack()[1][1]))


def is_frozen_as_exe() -> bool:
    """Check if script is frozen as an exe using PyInstaller.

    Eli (12/20/19) cannot figure out how to mock sys, so this is not
    unit tested.
    """
    return hasattr(sys, "_MEIPASS")


def get_path_to_frozen_bundle() -> str:  # pragma: no cover
    """Return path to the running bundle made by PyInstaller.

    Eli (12/20/19) cannot figure out how to mock sys, so this is not
    unit tested.
    """
    path_to_bundle = getattr(sys, "_MEIPASS")
    if not isinstance(path_to_bundle, str):
        raise NotImplementedError(
            "The _MEIPASS sys attribute should always be a string."
        )
    return path_to_bundle


def resource_path(relative_path: str, base_path: Optional[str] = None) -> str:
    """Get a path to a resource that works for development and frozen files.

    Args:
        relative_path: an additional path to append
        base_path: If left blank, this will default to the path to the file that called this function. When frozen as EXE, this will always be the path to the Pyinstaller directory.

    For use specifically with files compiled to windows EXE by pyinstaller.
    """
    if base_path is None:
        path_to_file_that_called_this_function = os.path.dirname(  # pylint: disable=invalid-name
            (inspect.stack()[1][1])
        )
        base_path = path_to_file_that_called_this_function
    if base_path == "":
        raise BlankAbsoluteResourcePathError()
    if is_frozen_as_exe():
        base_path = get_path_to_frozen_bundle()
    return os.path.join(base_path, relative_path)


def is_system_windows() -> bool:
    """Check if running on windows."""
    system_type = os.name
    return system_type == "nt"


# def raise_alarm_signal() -> None:
#     """Raise signal in a UNIX and Windows compatible manner.

#     Raises signal.SIGALRM which may not exist on windows, but is 14 as
#     an int. Raise it as fast as possible. In Python 3.8, raise_signal
#     may be a cross-platform option.
#     """
#     if is_system_windows():
#         # from https://stackoverflow.com/questions/14457723/can-i-raise-a-signal-from-python
#         ucrtbase = ctypes.CDLL("ucrtbase")
#         c_raise = ucrtbase["raise"]
#         c_raise(14)
#     else:
#         signal.setitimer(signal.ITIMER_REAL, 0.001)


def create_directory_if_not_exists(the_dir: str) -> None:
    if not os.path.exists(the_dir):
        os.makedirs(the_dir)


def get_formatted_stack_trace(e: Exception) -> str:
    # format the stack trace (Eli 2/7/20 couldn't figure out a way to get the stack trace from the exception itself once it had passed back into the main process, so need to grab it explicitly here) https://stackoverflow.com/questions/4564559/get-exception-description-and-stack-trace-which-caused-an-exception-all-as-a-st
    stack = traceback.extract_stack()[:-3] + traceback.extract_tb(
        e.__traceback__
    )  # add limit=??
    pretty = traceback.format_list(stack)
    formatted_stack_trace = "".join(pretty) + "\n  {} {}".format(e.__class__, e)
    return formatted_stack_trace


def print_exception(the_exception: Exception, call_id: Union[UUID, str]) -> None:
    print_warning_msg = "IMPORTANT: This fatal error message is being printed to the console before attempting to be logged. Confirm it is in the log file before closing the console. Screenshot or copy the console to save the error if it is not in the log!"
    stack_trace = get_formatted_stack_trace(the_exception)
    msg = f"{print_warning_msg}\nID of call to print: {call_id}\n{stack_trace}"
    print(msg)  # allow-print
